'''
This module primarily conatins a the construct_funcmap
function. The function constructs the dictionary that contains 
the console cases. The dictionary keys consists of a range of integers
and the values are the cases (which are functions). 
'''

from inspect import getmembers, isfunction, getmodule, unwrap
from types import ModuleType
from typing import Union, Callable
from collections import Iterable

def _docstring_firstline(func: Callable):
    '''Get first line of docstring of func'''
    # Unwrap in case the function is wrapped
    func = unwrap(func)
    if func.__doc__ is None:
        raise NotImplementedError(f'Missing docstring in function {func}')
    return func.__doc__.strip().split('\n')[0]

def print_funcmap(func_map: dict):
    '''
    Prints a func_map dictionary

    Items should be tuples with first elements as descriptions
    and second elements as function objects
    '''
    for key, tup in func_map.items():
        print(key + '.', tup[0])  

def __get_module_cases(module):
    # Get all functions defined in module
    f_ = lambda f: True if isfunction(f) and getmodule(f) == module else False
    funcs = getmembers(module, f_) 
    # getmembers returns a tuple with the func names as first element 
    # and function object as second

    # unpack dat shit yo
    funcs = [f[1] for f in funcs]
    return funcs

def construct_funcmap(cases: Union[ModuleType, list], other_cases: list=None, 
                      decorator: Callable=None):
    '''
    Parameters
    ------------
    cases: if given a module: module containing functions that serves as 
           cases a user can pick from terminal interface. the module should
           not implement any other functions. 
    
           if given a list: will simply use function in list as cases.

           First line of docstring becomes case description
           ALL CASES MUST CONTAIN DOCSTRINGS

    other_cases: optional, list of other additional cases (functions).
                 These functions will simply be appended to the end of 
                 an already existing list. 

                 IMPORTANT: All case functions must have docstrings 

    decorator: optional, a decorator to decorate all case functions 

    Returns
    --------
    Returns dictionary to be used in console interface

    Keys of dictionary are enumeration 1, 2, 3 ... for interface.
    Each item is a tuple with first element as description of case, 
    second element is the function itself:
    ('Scrape OSEBX', function object)
    '''
    if type(cases) == ModuleType:
        funcs = __get_module_cases(cases)
    elif type(cases) == list:
        funcs = cases
    else:
        raise TypeError('Unsupported type for cases container')

    # Append other functions if specified
    if other_cases is not None:
        funcs = funcs + other_cases

    func_map = {}

    if decorator is not None:
        for i, func in enumerate(funcs, start=1):
            func_map[str(i)] = (_docstring_firstline(func), decorator(func))
    else:
        for i, func in enumerate(funcs, start=1):
            func_map[str(i)] = (_docstring_firstline(func), func)

    return func_map


'''
Contains the command line interface (CLI) class, along its factory function:
menu()
'''

from time import sleep
import pypatconsole.consolestrings as strings
import pypatconsole.consoleconfig as ccng 
from pypatconsole.funcmap import (construct_funcmap, print_funcmap, 
                                 _docstring_firstline)
from pypatconsole.consolecommon import (clear_screen, input_splitter, 
                                        list_local_cases, print_help)
from typing import Union, Callable
from inspect import getfullargspec, unwrap, signature
from types import ModuleType

def logo_title(title: str):
    '''Prints logo title'''
    print("{:-^40s}".format(title))

def show_cases(funcmap: dict, title=strings.LOGO_TITLE):
    '''Prints function map prettily with a given title'''
    logo_title(title)
    print_funcmap(funcmap)

def enter_prompt(msg: str=strings.ENTER_PROMPT):
    '''Prints enter prompt message and than returns input()'''
    print(msg, end=': ')
    return input()

def exit_program(*args, **kwargs):
    '''
    Exit program
    '''
    print(strings.EXIT_MSG)
    sleep(ccng.MSG_WAIT_TIME)
    clear_screen()
    exit()

__SPECIAL_ARG_CASES = {
    str: lambda x: str(x.strip("\"\'")), # Removes outer " or ' characters
    tuple:eval, 
    list:eval, 
    dict:eval,
    set:eval
}

def _handle_arglist(func, arglist):
    '''
    Handles list of strings that are the arguments 
    The function turns the strings from the list into 
    their designated types (found from function signature).
    '''
    # Unwrap in case the function is wrapped
    func = unwrap(func)
    argsspec = getfullargspec(func)
    args = argsspec.args
    argtypes = argsspec.annotations

    if len(args) > len(argtypes):
        raise NotImplementedError(f'Missing typehints in {unwrap(func)}')

    if len(arglist) > len(args):
        raise ValueError(f'Got too many arguments, should be {len(args)}, but got {len(arglist)}')

    # Special proceedures for special classes
    typed_arglist = [] 
    for arg, type_ in zip(arglist, argtypes.values()):
        if type_ in __SPECIAL_ARG_CASES:
            typed_arglist.append(__SPECIAL_ARG_CASES[type_](arg))
        else:
            typed_arglist.append(type_(arg))

    return typed_arglist

def _error_info(error, func):
    print(strings.ERROR_INDICATOR)
    print(f'Selected case: "{_docstring_firstline(func)}"')
    print(error)
    print(f'Function signature: {signature(func)}')
    print()
    print(strings.INPUT_WAIT_PROMPT_MSG)
    input()

class CLI:
    '''
    Command Line Interface class 
    '''
    def __init__(self, cases, title: str=strings.LOGO_TITLE, 
                 blank_proceedure: Union[str, Callable]='return', 
                 decorator: Callable=None):
        '''
        Input
        -----
        cases: 

        if given a module: module containing functions that serves as cases a 
        user can pick from terminal interface. the module should not implement
        any other functions. 
        
        if given a list: will simply use function in list as cases.

        First line of docstring becomes case description
        ALL CASES MUST CONTAIN DOCSTRINGS

        title: String to print over alternatives 

        blank_proceedure: What to do when given blank input (defaults to 
                          stopping current view (without exiting))
        '''
        self.funcmap = construct_funcmap(cases, decorator=decorator)
        self.title = title

        if blank_proceedure == 'return':
            self.blank_hint = strings.INPUT_BLANK_HINT_RETURN
            self.blank_proceedure = self.__return_to_parent
        elif blank_proceedure is 'pass':
            self.blank_hint = strings.INPUT_BLANK_HINT_PASS
            self.blank_proceedure = self.__pass
        elif blank_proceedure == 'exit':
            self.blank_hint = strings.INPUT_BLANK_EXIT
            self.blank_proceedure = exit_program
        else:
            self.blank_proceedure = blank_proceedure

        # Special options
        self.special_cases = {
            '..':self.__return_to_parent,
            'q':exit_program,
            'h':print_help
        }

    def __return_to_parent(self):
        self.active = False
    
    def __pass(self):
        pass

    def _handle_case(self, casefunc, inputlist):
        try:
            if inputlist:
                # Raises TypeError if wrong number of arguments 
                casefunc(*_handle_arglist(casefunc, inputlist))
            else:
                # Will raise TypeError if casefunc() actually 
                # requires arguments
                casefunc()
        except TypeError as e:
            _error_info(e, casefunc)

    def run(self):
        '''
        Main function for console interface
        '''
        self.active = True
        try:
            while self.active:
                clear_screen()
                show_cases(self.funcmap, self.title)

                # Get key to func map
                print()
                print(self.blank_hint)
                inputstring = enter_prompt(strings.ENTER_PROMPT)

                # Pressing enter without specifying enables if test
                clear_screen()
                if not inputstring:
                    self.blank_proceedure()
                    continue

                # Tokenize input
                inputlist = input_splitter(inputstring)
                # Get case
                case = inputlist.pop(0)

                if case in self.funcmap:
                    # Obtain case function from funcmap and 
                    # calls said function. Recall that items are 
                    # (description, function), hence the [1]
                    casefunc = self.funcmap[case][1]
                    self._handle_case(casefunc, inputlist)
                elif case in self.special_cases:
                    # Items in special_cases are not tuples, but the 
                    # actual functions, so no need to do [1]
                    casefunc = self.special_cases[case]
                    self._handle_case(casefunc, inputlist)
                else:
                    print(strings.INVALID_TERMINAL_INPUT_MSG)
                    sleep(ccng.MSG_WAIT_TIME)

        except KeyboardInterrupt:
            # Ensures proper exit when Kbinterrupt
            exit_program()
    
def menu(cases: Union[list, dict, ModuleType], title: str=' Title ', 
                blank_proceedure: Union[str, Callable]='return', 
                decorator: Callable=None, run: bool=True, main: bool=False):
    '''
    Factory function for the CLI class. This function initializes a menu. 

    Parameters
    ------------
    cases: Can be output of locals() (a dictionary) from the scope of the cases 
           
           Or a list functions 

           Or a module containing the case functions

    title: title of menu

    blank_proceedure: What to do the when given blank input. Can be user defined
                      function, or it can be a string. Available string options
                      are:
                      
                      'return', will return to parent menu, if you are at main
                      menu, this will exit the program

                      'pass', does nothing. This should only be used for the 
                      main menu

                      'exit', exits the program

    decorator: Decorator for case functions

    run: To invoke .run() method on CLI object or not.

    main: Tells the function whether or not the menu is the main menu (i.e. the 
          first menu) or not. This is very import to specify when you give a 
          dictionary (e.g. the return value of locals()).

    Returns
    --------
    CLI (Command Line Interface) object. Use .run() method to activate menu. 
    '''
    if type(cases) == list:
        cases_to_send = cases
    elif type(cases) == dict:
        cases_to_send = list_local_cases(cases)
        # If this menu is the first menu initialized, and is given the locally
        # defined functions, then must filter the functions that are defined 
        # in __main__
        if main:
            cases_to_send =\
                [c for c in cases_to_send if c.__module__ == '__main__']
            blank_proceedure='pass'

    elif type(cases) == ModuleType:
        cases_to_send = cases
    else:
        raise TypeError('Invalid type')

    cli = CLI(cases=cases_to_send, title=title, 
              blank_proceedure=blank_proceedure, decorator=decorator)
    if run:
        cli.run()
    
    return cli



PURPLE = '\033[95m'
CYAN = '\033[96m'
DARKCYAN = '\033[36m'
BLUE = '\033[94m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
RED = '\033[91m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'
END = '\033[0m'

START_MSG = 'Starting console interface'
LOGO_TITLE = ' Main menu '


ENTER_PROMPT = 'Input'

EXIT_MSG = 'Exiting program'
INVALID_TERMINAL_INPUT_MSG = 'Invalid input'
ERROR_INDICATOR = BOLD + RED + f"{' ERROR ':#^40}" + END

INPUT_BLANK_HINT_RETURN = 'Enter blank to return to parent menu'
INPUT_BLANK_HINT_PASS = ''
INPUT_BLANK_HINT_EXIT = 'Enter blank to exit program'

INPUT_WAIT_PROMPT_MSG = 'Press enter to continue'



'''
Not used in current version
'''

from time import sleep
import consoleconfig as ccng

def case_decorator(func):
    '''Decorator to enforce commmon behavior for cases'''
    
    def case_wrapper(*args, **kwargs):
        '''Feel free to customize wrapper'''
        retobj = func(*args, **kwargs)
        sleep(ccng.CASE_EXIT_WAIT_TIME) 
        return retobj

    # "Inherit" docstring (not really neccessary, but kinda nice to have)
    case_wrapper.__doc__ = func.__doc__
    # This is necessary in order to unwrap function from inspect module
    case_wrapper.__wrapped__ = func
    return case_wrapper



'''
Common stuff for console stuff
'''
import os 
from inspect import isfunction
from typing import Union, Callable
from types import ModuleType
from shlex import shlex

__CLEAR_COMMAND = 'cls' if os.name == 'nt' else 'clear'
def clear_screen():
    '''Obvious'''
    os.system(__CLEAR_COMMAND)

def list_local_cases(locals_: dict, main: bool=False):
    '''
    Parameters
    -------------
    locals_: return value of locals()
    
    Returns a list of functions sorted alphabetically by function names. 
    '''
    name_func_pairs = sorted(list(locals_.items()), key=lambda x: x[0])
    return [pairs[1] for pairs in name_func_pairs if isfunction(pairs[1])]

def input_splitter(argstring: str):
    '''
    Split string
    '''
    # Doing it this way instead of using shlex.split will
    # not remove quote symbols in dicts and lists and whatever 
    shlexysmexy = shlex(argstring)
    shlexysmexy.whitespace_split=True
    return [token for token in shlexysmexy]

def print_help(*args, **kwargs):
    print(
        '''
        To exit console: Enter q or press Ctrl+c

        To return to parent menu: Enter blank (press enter without giving input)
                                  or enter '..'. If you are in main menu, this 
                                  will exit the program as well. 

        Press enter to exit help screen
        '''
    )
    input()

